#ifndef _apachetop_h
#define _apachetop_h

#include <errno.h>
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <time.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <curses.h>

#include "arch.h"

#include "log.h"
#include "hashes.h"
#include "ohtbl.h"
#include "map.h"
#include "circle.h"
#include "hits_circle.h"
#include "timed_circle.h"

#define getMIN(a,b) (a < b ? a : b)
#define getMAX(a,b) (a > b ? a : b)

struct config {
#define SORT_REQCOUNT 1
#define SORT_REQPERSEC 2
#define SORT_BYTECOUNT 3
#define SORT_BYTESPERSEC 4
	short sort;
	short refresh_delay;

	short selected_item; /* position our marker is at */
	unsigned int selected_item_pos; /* which urlpos it relates to */

	char *logfile;
	/* types defined in log.h */
	short logtype;

#define TIMED_CIRCLE 'T'
#define HITS_CIRCLE 'H'
	int circle_mode;

	int circle_size;

#define DISPLAY_URLS 1
#define DISPLAY_IPS 2
#define DISPLAY_REFS 3
	short display_mode;

	/* if nonzero, show referrer stats for the given url_pos */
	int display_referrers;

	/* url munging */
	bool keep_querystring;
	bool lowercase_urls;
	unsigned short keep_segments;
	bool preserve_ref_protocol;
};

struct peritem_stat {
	unsigned int item;
	long long reqcount;
	long long bytecount;
	time_t first, last, timespan;
	float rps, kbps;
};

struct gstat {
	time_t start; /* when did we start */

	/* space for 1xx-5xx return codes */
	struct r_code {
		long long bytecount;
		long long reqcount;
	} r_codes[6];

	/* space for counting global hits and bytes per sec */
	struct global {
		long long bytecount;
		long long reqcount;
		time_t first, last;
	} alltime;

};


#define JAN 281
#define FEB 269
#define MAR 288
#define APR 291
#define MAY 295
#define JUN 301
#define JUL 299
#define AUG 285
#define SEP 296
#define OCT 294
#define NOV 307
#define DEC 268

#define DEFAULT_LOGFILE "/var/httpd/apache_log"
#define DEFAULT_CIRCLE_SIZE 30
#define DEFAULT_CIRCLE_MODE TIMED_CIRCLE

/* if the layout of the display changes, these need updating */
#define COLS_RESERVED 25
#define LINES_RESERVED 7

int dprintf(const char *fmt, ...);

void display();
void displayReferrers();
void showDetail(struct peritem_stat *item_ptr, int disp);

int reqcount_cmp(const void *a, const void *b);
int rps_cmp(const void *a, const void *b);
int bytes_cmp(const void *a, const void *b);
int bps_cmp(const void *a, const void *b);

float readableNum(long long num, char *suffix);


int readKey(int ch);
void usage(char *cmd);

#endif
